# Assuming img is your image tensor of shape [B, C, H, W]
def crop_image_border(img, crop_border):
    """
    Crop the borders of an image.

    Parameters:
    - img: A tensor representing the image, shape [B, C, H, W].
    - crop_border: The number of pixels to crop from each border (top, bottom, left, right).

    Returns:
    - Cropped image tensor.
    """
    B, C, H, W = img.shape
    return img[:, :, crop_border:-crop_border, crop_border:-crop_border]




from torch import nn
from niiv.decoder.sine_layer import SineLayer

class FieldSiren(nn.Module):
    network: nn.Sequential

    def __init__(
        self,
        d_coordinate: int,
        n_layers: int,
        n_neurons: int,
        d_out: int,
    ) -> None:
        """Set up a SIREN network using the sine layers"""
        super().__init__()

        layers = []
        layers.append(SineLayer(d_coordinate, n_neurons))
        for i in range(n_layers - 1):
            layers.append(SineLayer(n_neurons, n_neurons))
        layers.append(nn.Linear(n_neurons, d_out))

        self.model = nn.Sequential(*layers)

    def forward(self, coordinates):
        """Evaluate the MLP at the specified coordinates."""
        return self.model(coordinates)



# Based on Wire (CVPR 2024)
# Code adapted from: https://github.com/vishwa91/wire/blob/main/modules/wire.py

import torch 
import torch.nn as nn
import numpy as np

from niiv.decoder.wire import ComplexGaborLayer

class Gabor(nn.Module):
    def __init__(self, in_features, hidden_features, 
                 hidden_layers, 
                 out_features,
                 first_omega_0=10.0, hidden_omega_0=10.0, scale=5.0):
        super().__init__()
        
        # All results in the paper were with the default complex 'gabor' nonlinearity
        self.nonlin = ComplexGaborLayer
        
        # Since complex numbers are two real numbers, reduce the number of 
        # hidden parameters by 2
        hidden_features = int(hidden_features/np.sqrt(2))
        dtype = torch.cfloat
        self.complex = True
        self.wavelet = 'gabor'
            
        self.net = []
        self.net.append(self.nonlin(in_features,
                                    hidden_features, 
                                    omega0=first_omega_0,
                                    sigma0=scale,
                                    is_first=True,
                                    trainable=False))

        for i in range(hidden_layers):
            self.net.append(self.nonlin(hidden_features,
                                        hidden_features, 
                                        omega0=hidden_omega_0,
                                        sigma0=scale))

        final_linear = nn.Linear(hidden_features,
                                 out_features,
                                 dtype=dtype)            
        self.net.append(final_linear)
        
        self.net = nn.Sequential(*self.net)
    
    def forward(self, coords):
        output = self.net(coords)
        
        if self.wavelet == 'gabor':
            return output.real
         
        return output


import numpy as np
import cv2
import tqdm

a = np.load(
    "/home/jakobtroidl/Desktop/neural-volumes/data/ista/20230716_ExPID114_17.50.53.npy"
)

out_path = '/home/jakobtroidl/Desktop/neural-volumes/data/ista/20230716_ExPID114_17.50.53_clahe.npy'

print(a.shape)
print(a.dtype)
print(a.min(), a.max())

clahe = cv2.createCLAHE(clipLimit=3.0, tileGridSize=(127,127))
transformed_volume = np.zeros_like(a)

# Iterate over each slice along the first dimension (Z-axis)
for i in tqdm(range(a.shape[0])):
    # Assuming the transform is CLAHE for demonstration; replace with your transform.
    slice = clahe.apply(a[i, :, :])
    transformed_volume[i, :, :] = slice

np.save(out_path, transformed_volume)


import torch.nn as nn
import torch


class PositionalEncoding(nn.Module):
    def __init__(self, num_octaves: int):
        super().__init__()
        self.num_octaves = num_octaves

    def forward(self, samples):
        """Separately encode each channel using a positional encoding. The lowest
        frequency should be 2 * torch.pi, and each frequency thereafter should be
        double the previous frequency. For each frequency, you should encode the input
        signal using both sine and cosine.
        """

        if len(samples.shape) == 1:
            samples = samples.unsqueeze(0)

        freq = (
            2 ** torch.arange(1, self.num_octaves + 1, dtype=torch.float32) * torch.pi
        ).to(samples.device)

        freq = freq.view(1, 1, -1)  # prepare for broadcasting
        samples = samples[..., None]  # prepare for broadcasting

        phase = samples * freq

        phase_shape = list(phase.shape)
        phase_transformed_shape = phase_shape[:-2] + [phase_shape[-2] * phase_shape[-1]]
        sin = torch.sin(phase).reshape(*phase_transformed_shape)
        cos = torch.cos(phase).reshape(*phase_transformed_shape)

        output = torch.stack([sin, cos], dim=-1)

        shape = list(output.shape)
        transformed_shape = shape[:-2] + [output.shape[-1] * output.shape[-2]]
        output = output.reshape(*transformed_shape)

        return output

    def d_out(self, dimensionality: int):
        return 2 * self.num_octaves * dimensionality



import torch
from torch.fft import fft2, fftshift, ifft2, ifftshift
from torchmetrics.image import PeakSignalNoiseRatio, StructuralSimilarityIndexMeasure

def write_metrics_string(metrics, names):
    out = ""
    for i, name in enumerate(names):
        out += f"{name}: {metrics[:, i].mean().item():.2f}, "
    return out


def compute_all_metrics(pred, gt, im_max=1.0, im_size=128):
    image_psnr = compute_metric(ImagePSNR(), pred, gt)
    image_ssim = compute_metric(ImageSSIM(), pred, gt)
    clipped_fourier_psnr = compute_metric(ClippedFourierPSNR(im_max, im_size), pred, gt)

    return torch.stack((image_psnr, image_ssim, clipped_fourier_psnr), dim=1)


def compute_metric(metric, pred, gt):
    return metric(pred, gt)

class LowPassFilter(torch.nn.Module):
    def __init__(self, rows, cols) -> None:
        super(LowPassFilter, self).__init__()
        self.rows = rows
        self.cols = cols
    
    def forward(self, radius):
        crow, ccol = self.rows // 2 , self.cols // 2
        low_pass = torch.zeros((self.rows, self.cols), dtype=torch.uint8)
        x = torch.arange(0, self.cols).unsqueeze(0).expand(self.rows, self.cols)
        y = torch.arange(0, self.rows).unsqueeze(1).expand(self.rows, self.cols)
        mask_area = (x - ccol) ** 2 + (y - crow) ** 2 <= radius**2

        delta = radius // 2
        mid = self.cols // 2
        start = max(mid - delta, 0)
        end = min(mid + delta, self.cols)
        # low_pass[:, start:end ] = 1
        low_pass[mask_area] = 1

        return low_pass.to("cuda")

class MultiClippedFourierPSNR(torch.nn.Module):
    def __init__(self, n_thresholds=64, im_max=1.0, im_size=128) -> None:
        super(MultiClippedFourierPSNR, self).__init__()
        self.cf_psnr = ClippedFourierPSNR(im_max, im_size)
        end = (im_size * torch.sqrt(torch.tensor(2))) // 2
        # end = 128
        t = torch.linspace(2, end, n_thresholds)
        self.thresholds = torch.round(t).int()

    def thresholds(self):
        return self.thresholds.tolist()
    
    def forward(self, prediction, gt):
        cf_psnr = torch.zeros((prediction.shape[0], self.thresholds.shape[0]), dtype=torch.float32).to("cuda")
        for i, threshold in enumerate(self.thresholds):
            cf_psnr[:, i] = self.cf_psnr(prediction, gt, threshold)
        return cf_psnr

class FourierDenoiser(torch.nn.Module):
    def __init__(self, threshold=25):
        super(FourierDenoiser, self).__init__()
        self.filter_threshold = threshold
    
    def forward(self, x):
        f_transform = fft2(x)
        f_shifted = fftshift(f_transform)
        low_pass = LowPassFilter(x.shape[-2], x.shape[-1])(self.filter_threshold)
        f_filtered = f_shifted * low_pass
        return torch.abs(ifft2(ifftshift(f_filtered)))


class ClippedFourierPSNR(torch.nn.Module):
    def __init__(self, im_max=1.0, im_size=128):
        super(ClippedFourierPSNR, self).__init__()
        self.im_max = torch.tensor(im_max)
        self.low_pass = LowPassFilter(im_size, im_size)

    def forward(self, prediction, gt, threshold=25):
        n_pixels = prediction.shape[-1] * prediction.shape[-2]
        pred_fourier = fftshift(fft2(prediction)) * self.low_pass(threshold)
        gt_fourier = fftshift(fft2(gt)) * self.low_pass(threshold)
        sq_error = torch.abs(pred_fourier - gt_fourier) ** 2
        sum_sq_error = torch.sum(sq_error, dim=[-3, -2, -1])
        cf_psnr = 20 * torch.log10(self.im_max * n_pixels) - 10 * torch.log10(sum_sq_error)
        return cf_psnr
    
class ImagePSNR(torch.nn.Module):
    def __init__(self, im_max=1.0):
        super(ImagePSNR, self).__init__()
        self.psnr_metric = PeakSignalNoiseRatio(data_range=im_max, reduction='none', dim=[1, 2, 3]).cuda()

    def forward(self, prediction, gt):
        return self.psnr_metric(prediction, gt)


class ImageSSIM(torch.nn.Module):
    def __init__(self, im_max=1.0):
        super(ImageSSIM, self).__init__()
        self.ssim_metric = StructuralSimilarityIndexMeasure(data_range=im_max, reduction='none').cuda()

    def forward(self, prediction, gt):
        return self.ssim_metric(prediction, gt)


import torch
import os

from pytorch_msssim import ms_ssim
class Averager():
    def __init__(self):
        self.n = 0.0
        self.v = 0.0

    def add(self, v, n=1.0):
        self.v = (self.v * self.n + v * n) / (self.n + n)
        self.n += n

    def item(self):
        return self.v

def msssim_fn(output, target):
    if output.size(-2) >= 160:
        msssim = ms_ssim(output.float().detach(), target.detach(), data_range=1, size_average=True)
    else:
        msssim = torch.tensor(0).to(output.device)

    return msssim

def get_n_params(model):
    pp=0
    for p in list(model.parameters()):
        nn=1
        for s in list(p.size()):
            nn = nn*s
        pp += nn
    return pp
    
def cond_mkdir(path):
    if not os.path.exists(path):
        os.makedirs(path)

def write_volume_summary(volume_dataset, model, model_input, gt, writer, total_steps, prefix='train_'):
    print("write_volume_summary")
    PSNR = 0.0 # TODO
    return PSNR

def min_max_summary(name, tensor, writer, total_steps):
    writer.add_scalar(name + '_min', tensor.min().detach().cpu().numpy(), total_steps)
    writer.add_scalar(name + '_max', tensor.max().detach().cpu().numpy(), total_steps)

# from LIIF code base (Chen. 2021)
def make_coord(shape, ranges=None, flatten=True):
    """ Make coordinates at grid centers.
    """
    coord_seqs = []
    for i, n in enumerate(shape):
        if ranges is None:
            v0, v1 = -1, 1
        else:
            v0, v1 = ranges[i]
        r = (v1 - v0) / (2 * n)
        seq = v0 + r + (2 * r) * torch.arange(n).float()
        # seq = v0 + (2 * r) * torch.arange(n).float()
        coord_seqs.append(seq)
    ret = torch.stack(torch.meshgrid(*coord_seqs), dim=-1)
    if flatten:
        ret = ret.view(-1, ret.shape[-1])
    return ret

def exclude_max_min(list):
    sorted_list = sorted(list)
    return sorted_list[1:-1]








# Residual Dense Network for Image Super-Resolution
# https://arxiv.org/abs/1802.08797
# modified from: https://github.com/jaewon-lee-b/lte/blob/main/models/rdn.py

from argparse import Namespace

import torch
import torch.nn as nn


class RDB_Conv(nn.Module):
    def __init__(self, inChannels, growRate, kSize=3):
        super(RDB_Conv, self).__init__()
        Cin = inChannels
        G  = growRate
        self.conv = nn.Sequential(*[
            nn.Conv2d(Cin, G, kSize, padding=(kSize-1)//2, stride=1),
            nn.ReLU()
        ])

    def forward(self, x):
        out = self.conv(x)
        return torch.cat((x, out), 1)

class RDB(nn.Module):
    def __init__(self, growRate0, growRate, nConvLayers, kSize=3):
        super(RDB, self).__init__()
        G0 = growRate0
        G  = growRate
        C  = nConvLayers

        convs = []
        for c in range(C):
            convs.append(RDB_Conv(G0 + c*G, G))
        self.convs = nn.Sequential(*convs)

        # Local Feature Fusion
        self.LFF = nn.Conv2d(G0 + C*G, G0, 1, padding=0, stride=1)

    def forward(self, x):
        return self.LFF(self.convs(x)) + x

class RDN(nn.Module):
    def __init__(self, args):
        super(RDN, self).__init__()
        self.args = args
        r = args.scale[0]
        G0 = args.G0
        kSize = args.RDNkSize

        # number of RDB blocks, conv layers, out channels
        self.D, C, G = {
            'A': (20, 6, 32),
            'B': (16, 8, 64),
        }[args.RDNconfig]

        # Shallow feature extraction net
        self.SFENet1 = nn.Conv2d(args.n_colors, G0, kSize, padding=(kSize-1)//2, stride=1)
        self.SFENet2 = nn.Conv2d(G0, G0, kSize, padding=(kSize-1)//2, stride=1)

        # Redidual dense blocks and dense feature fusion
        self.RDBs = nn.ModuleList()
        for i in range(self.D):
            self.RDBs.append(
                RDB(growRate0 = G0, growRate = G, nConvLayers = C)
            )

        # Global Feature Fusion
        self.GFF = nn.Sequential(*[
            nn.Conv2d(self.D * G0, G0, 1, padding=0, stride=1),
            nn.Conv2d(G0, G0, kSize, padding=(kSize-1)//2, stride=1)
        ])

        if args.no_upsampling:
            self.out_dim = G0
        else:
            self.out_dim = args.n_colors
            # Up-sampling net
            if r == 2 or r == 3:
                self.UPNet = nn.Sequential(*[
                    nn.Conv2d(G0, G * r * r, kSize, padding=(kSize-1)//2, stride=1),
                    nn.PixelShuffle(r),
                    nn.Conv2d(G, args.n_colors, kSize, padding=(kSize-1)//2, stride=1)
                ])
            elif r == 4:
                self.UPNet = nn.Sequential(*[
                    nn.Conv2d(G0, G * 4, kSize, padding=(kSize-1)//2, stride=1),
                    nn.PixelShuffle(2),
                    nn.Conv2d(G, G * 4, kSize, padding=(kSize-1)//2, stride=1),
                    nn.PixelShuffle(2),
                    nn.Conv2d(G, args.n_colors, kSize, padding=(kSize-1)//2, stride=1)
                ])
            else:
                raise ValueError("scale must be 2 or 3 or 4.")

    def forward(self, x):
        f__1 = self.SFENet1(x)
        x  = self.SFENet2(f__1)

        RDBs_out = []
        for i in range(self.D):
            x = self.RDBs[i](x)
            RDBs_out.append(x)

        x = self.GFF(torch.cat(RDBs_out,1))
        x += f__1

        if self.args.no_upsampling:
            return x
        else:
            return self.UPNet(x)


def make_rdn(G0=64, RDNkSize=3, RDNconfig='B',
             scale=2, no_upsampling=True):
    args = Namespace()
    args.G0 = G0
    args.RDNkSize = RDNkSize
    args.RDNconfig = RDNconfig

    args.scale = [scale]
    args.no_upsampling = no_upsampling

    args.n_colors = 1
    return RDN(args)



import torch
from torchmetrics.image import StructuralSimilarityIndexMeasure

class SSIM_Loss(torch.nn.Module):
    def __init__(self, range=1.0):
        super(SSIM_Loss, self).__init__()
        self.ssim = StructuralSimilarityIndexMeasure(data_range=range).to('cuda')

    def forward(self, prediction, gt):
        return 1.0 - self.ssim(prediction, gt)


def image_mse(mask, model_output, gt):
    model_output = model_output.squeeze()
    gt = gt.squeeze()
    if mask is None:
        return {'img_loss': ((model_output- gt) ** 2).mean()}
    else:
        return {'img_loss': (mask * (model_output - gt) ** 2).mean()}
    
def image_l1(model_output, gt):
    model_output = model_output.squeeze()
    gt = gt.squeeze()
    return torch.abs(model_output - gt).mean()


def charbonnier_loss(prediction, gt, epsilon=1e-6):
    diff = prediction - gt
    loss = torch.sum(torch.sqrt(diff * diff + epsilon))
    return loss


def multi_frame_loss(model_output, gt, overlap, overlap_gt):

    weight_img_loss = 1.0
    weight_overlap_loss = 0.0
    
    image_loss = image_l1(model_output, gt)
    overlap_loss = image_l1(overlap, overlap_gt)

    total_loss = weight_img_loss * image_loss + weight_overlap_loss * overlap_loss
    
    return [total_loss, image_loss, overlap_loss]



import numpy as np
import torch
from torch import nn


class SineLayer(nn.Module):
    def __init__(
        self,
        d_in: int,
        d_out: int,
        bias: bool = True,
        is_first: bool = True,
        omega_0: float = 30.0,
    ):
        super().__init__()
        self.omega_0 = omega_0
        self.is_first = is_first
        self.d_in = d_in
        self.linear = nn.Linear(d_in, d_out, bias=bias)
        self.init_weights()

    def init_weights(self):
        with torch.no_grad():
            if self.is_first:
                self.linear.weight.uniform_(-1 / self.d_in, 1 / self.d_in)
            else:
                self.linear.weight.uniform_(
                    -np.sqrt(6 / self.d_in) / self.omega_0,
                    np.sqrt(6 / self.d_in) / self.omega_0,
                )

    def forward(self, input):
        return torch.sin(self.omega_0 * self.linear(input))


import torch.nn as nn

class MLP(nn.Module):

    def __init__(self, in_dim, out_dim, n_hidden, n_neurons):
        super().__init__()

        layers = []
        self.in_dim = in_dim
        
        self.out_dim = out_dim
        lastv = self.in_dim

        for i in range(n_hidden):
            layers.append(nn.Linear(lastv, n_neurons))
            layers.append(nn.ReLU())
            lastv = n_neurons
        layers.append(nn.Linear(lastv, out_dim))
        self.layers = nn.Sequential(*layers)

    def forward(self, x):
        shape = x.shape[:-1]
        x = self.layers(x.view(-1, x.shape[-1]))
        return x.view(*shape, -1)



# modified from: https://github.com/thstkdgus35/EDSR-PyTorch

import math
from argparse import Namespace

import torch
import torch.nn as nn
# import torch.nn.functional as F

def default_conv(in_channels, out_channels, kernel_size, bias=True):
    return nn.Conv2d(
        in_channels, out_channels, kernel_size,
        padding=(kernel_size//2), bias=bias)

class MeanShift(nn.Conv2d):
    def __init__(
        self, rgb_range,
        rgb_mean=(0.4488, 0.4371, 0.4040), rgb_std=(1.0, 1.0, 1.0), sign=-1):

        super(MeanShift, self).__init__(3, 3, kernel_size=1)
        std = torch.Tensor(rgb_std)
        self.weight.data = torch.eye(3).view(3, 3, 1, 1) / std.view(3, 1, 1, 1)
        self.bias.data = sign * rgb_range * torch.Tensor(rgb_mean) / std
        for p in self.parameters():
            p.requires_grad = False

class ResBlock(nn.Module):
    def __init__(
        self, conv, n_feats, kernel_size,
        bias=True, bn=False, act=nn.ReLU(True), res_scale=1):

        super(ResBlock, self).__init__()
        m = []
        for i in range(2):
            m.append(conv(n_feats, n_feats, kernel_size, bias=bias))
            if bn:
                m.append(nn.BatchNorm2d(n_feats))
            if i == 0:
                m.append(act)

        self.body = nn.Sequential(*m)
        self.res_scale = res_scale

    def forward(self, x):
        res = self.body(x).mul(self.res_scale)
        res += x

        return res

class Upsampler(nn.Sequential):
    def __init__(self, conv, scale, n_feats, bn=False, act=False, bias=True):

        m = []
        if (scale & (scale - 1)) == 0:    # Is scale = 2^n?
            for _ in range(int(math.log(scale, 2))):
                m.append(conv(n_feats, 4 * n_feats, 3, bias))
                m.append(nn.PixelShuffle(2))
                if bn:
                    m.append(nn.BatchNorm3d(n_feats))
                if act == 'relu':
                    m.append(nn.ReLU(True))
                elif act == 'prelu':
                    m.append(nn.PReLU(n_feats))

        elif scale == 3:
            m.append(conv(n_feats, 9 * n_feats, 3, bias))
            m.append(nn.PixelShuffle(3))
            if bn:
                m.append(nn.BatchNorm3d(n_feats))
            if act == 'relu':
                m.append(nn.ReLU(True))
            elif act == 'prelu':
                m.append(nn.PReLU(n_feats))
        else:
            raise NotImplementedError

        super(Upsampler, self).__init__(*m)


class EDSR2D(nn.Module):
    def __init__(self, args=None, conv=default_conv):
        super(EDSR2D, self).__init__()
        if args is None:
            n_resblocks = 16
            n_feats = 64
            res_scale = 1
            n_colors = 1
        else:
            n_resblocks = args["n_resblocks"]
            n_feats = args["n_features"]
            n_colors = args["n_colors"]
            res_scale = args["res_scale"]
        kernel_size = 3
        act = nn.ReLU(True)

        # define head module
        m_head = [conv(n_colors, n_feats, kernel_size)]

        # define body module
        m_body = [
            ResBlock(
                conv, n_feats, kernel_size, act=act, res_scale=res_scale
            ) for _ in range(n_resblocks)
        ]
        m_body.append(conv(n_feats, n_feats, kernel_size))

        self.head = nn.Sequential(*m_head)
        self.body = nn.Sequential(*m_body)

        self.out_dim = n_feats

    def forward(self, x):
        x = self.head(x)
        res = self.body(x)
        res += x
        x = res
        return x

    def load_state_dict(self, state_dict, strict=True):
        own_state = self.state_dict()
        for name, param in state_dict.items():
            if name in own_state:
                if isinstance(param, nn.Parameter):
                    param = param.data
                try:
                    own_state[name].copy_(param)
                except Exception:
                    if name.find('tail') == -1:
                        raise RuntimeError('While copying the parameter named {}, '
                                           'whose dimensions in the model are {} and '
                                           'whose dimensions in the checkpoint are {}.'
                                           .format(name, own_state[name].size(), param.size()))
            elif strict:
                if name.find('tail') == -1:
                    raise KeyError('unexpected key "{}" in state_dict'
                                   .format(name))


# def make_edsr_baseline(n_resblocks=16, n_feats=64, res_scale=1,
#                        scale=2, no_upsampling=False, rgb_range=1):
#     args = Namespace()
#     args.n_resblocks = n_resblocks
#     args.n_feats = n_feats
#     args.res_scale = res_scale

#     args.scale = [scale]
#     args.no_upsampling = no_upsampling

#     args.rgb_range = rgb_range
#     args.n_colors = 1
#     return EDSR(args)


# def make_edsr(n_resblocks=32, n_feats=256, res_scale=0.1,
#               scale=2, no_upsampling=False, rgb_range=1):
#     args = Namespace()
#     args.n_resblocks = n_resblocks
#     args.n_feats = n_feats
#     args.res_scale = res_scale

#     args.scale = [scale]
#     args.no_upsampling = no_upsampling

#     args.rgb_range = rgb_range
#     args.n_colors = 1
#     return EDSR(args)



# # -----------------------------------------------------------------------------------
# # SwinIR: Image Restoration Using Swin Transformer, https://arxiv.org/abs/2108.10257
# # Originally Written by Ze Liu, Modified by Jingyun Liang.
# # ----------------------------------------------------------------------------------
# # modified from: https://github.com/jaewon-lee-b/lte/blob/main/models/swinir.py

# import math
# import torch
# import torch.nn as nn
# import torch.nn.functional as F
# import torch.utils.checkpoint as checkpoint
# from timm.models.layers import DropPath, to_2tuple, trunc_normal_

# from argparse import Namespace


# class Mlp(nn.Module):
#     def __init__(self, in_features, hidden_features=None, out_features=None, act_layer=nn.GELU, drop=0.):
#         super().__init__()
#         out_features = out_features or in_features
#         hidden_features = hidden_features or in_features
#         self.fc1 = nn.Linear(in_features, hidden_features)
#         self.act = act_layer()
#         self.fc2 = nn.Linear(hidden_features, out_features)
#         self.drop = nn.Dropout(drop)

#     def forward(self, x):
#         x = self.fc1(x)
#         x = self.act(x)
#         x = self.drop(x)
#         x = self.fc2(x)
#         x = self.drop(x)
#         return x


# def window_partition(x, window_size):
#     """
#     Args:
#         x: (B, H, W, C)
#         window_size (int): window size

#     Returns:
#         windows: (num_windows*B, window_size, window_size, C)
#     """
#     B, H, W, C = x.shape
#     x = x.view(B, H // window_size, window_size, W // window_size, window_size, C)
#     windows = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(-1, window_size, window_size, C)
#     return windows


# def window_reverse(windows, window_size, H, W):
#     """
#     Args:
#         windows: (num_windows*B, window_size, window_size, C)
#         window_size (int): Window size
#         H (int): Height of image
#         W (int): Width of image

#     Returns:
#         x: (B, H, W, C)
#     """
#     B = int(windows.shape[0] / (H * W / window_size / window_size))
#     x = windows.view(B, H // window_size, W // window_size, window_size, window_size, -1)
#     x = x.permute(0, 1, 3, 2, 4, 5).contiguous().view(B, H, W, -1)
#     return x


# class WindowAttention(nn.Module):
#     r""" Window based multi-head self attention (W-MSA) module with relative position bias.
#     It supports both of shifted and non-shifted window.

#     Args:
#         dim (int): Number of input channels.
#         window_size (tuple[int]): The height and width of the window.
#         num_heads (int): Number of attention heads.
#         qkv_bias (bool, optional):  If True, add a learnable bias to query, key, value. Default: True
#         qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set
#         attn_drop (float, optional): Dropout ratio of attention weight. Default: 0.0
#         proj_drop (float, optional): Dropout ratio of output. Default: 0.0
#     """

#     def __init__(self, dim, window_size, num_heads, qkv_bias=True, qk_scale=None, attn_drop=0., proj_drop=0.):

#         super().__init__()
#         self.dim = dim
#         self.window_size = window_size  # Wh, Ww
#         self.num_heads = num_heads
#         head_dim = dim // num_heads
#         self.scale = qk_scale or head_dim ** -0.5

#         # define a parameter table of relative position bias
#         self.relative_position_bias_table = nn.Parameter(
#             torch.zeros((2 * window_size[0] - 1) * (2 * window_size[1] - 1), num_heads))  # 2*Wh-1 * 2*Ww-1, nH

#         # get pair-wise relative position index for each token inside the window
#         coords_h = torch.arange(self.window_size[0])
#         coords_w = torch.arange(self.window_size[1])
#         coords = torch.stack(torch.meshgrid([coords_h, coords_w]))  # 2, Wh, Ww
#         coords_flatten = torch.flatten(coords, 1)  # 2, Wh*Ww
#         relative_coords = coords_flatten[:, :, None] - coords_flatten[:, None, :]  # 2, Wh*Ww, Wh*Ww
#         relative_coords = relative_coords.permute(1, 2, 0).contiguous()  # Wh*Ww, Wh*Ww, 2
#         relative_coords[:, :, 0] += self.window_size[0] - 1  # shift to start from 0
#         relative_coords[:, :, 1] += self.window_size[1] - 1
#         relative_coords[:, :, 0] *= 2 * self.window_size[1] - 1
#         relative_position_index = relative_coords.sum(-1)  # Wh*Ww, Wh*Ww
#         self.register_buffer("relative_position_index", relative_position_index)

#         self.qkv = nn.Linear(dim, dim * 3, bias=qkv_bias)
#         self.attn_drop = nn.Dropout(attn_drop)
#         self.proj = nn.Linear(dim, dim)

#         self.proj_drop = nn.Dropout(proj_drop)

#         trunc_normal_(self.relative_position_bias_table, std=.02)
#         self.softmax = nn.Softmax(dim=-1)

#     def forward(self, x, mask=None):
#         """
#         Args:
#             x: input features with shape of (num_windows*B, N, C)
#             mask: (0/-inf) mask with shape of (num_windows, Wh*Ww, Wh*Ww) or None
#         """
#         B_, N, C = x.shape
#         qkv = self.qkv(x).reshape(B_, N, 3, self.num_heads, C // self.num_heads).permute(2, 0, 3, 1, 4)
#         q, k, v = qkv[0], qkv[1], qkv[2]  # make torchscript happy (cannot use tensor as tuple)

#         q = q * self.scale
#         attn = (q @ k.transpose(-2, -1))

#         relative_position_bias = self.relative_position_bias_table[self.relative_position_index.view(-1)].view(
#             self.window_size[0] * self.window_size[1], self.window_size[0] * self.window_size[1], -1)  # Wh*Ww,Wh*Ww,nH
#         relative_position_bias = relative_position_bias.permute(2, 0, 1).contiguous()  # nH, Wh*Ww, Wh*Ww
#         attn = attn + relative_position_bias.unsqueeze(0)

#         if mask is not None:
#             nW = mask.shape[0]
#             attn = attn.view(B_ // nW, nW, self.num_heads, N, N) + mask.unsqueeze(1).unsqueeze(0)
#             attn = attn.view(-1, self.num_heads, N, N)
#             attn = self.softmax(attn)
#         else:
#             attn = self.softmax(attn)

#         attn = self.attn_drop(attn)

#         x = (attn @ v).transpose(1, 2).reshape(B_, N, C)
#         x = self.proj(x)
#         x = self.proj_drop(x)
#         return x

#     def extra_repr(self) -> str:
#         return f'dim={self.dim}, window_size={self.window_size}, num_heads={self.num_heads}'

#     def flops(self, N):
#         # calculate flops for 1 window with token length of N
#         flops = 0
#         # qkv = self.qkv(x)
#         flops += N * self.dim * 3 * self.dim
#         # attn = (q @ k.transpose(-2, -1))
#         flops += self.num_heads * N * (self.dim // self.num_heads) * N
#         #  x = (attn @ v)
#         flops += self.num_heads * N * N * (self.dim // self.num_heads)
#         # x = self.proj(x)
#         flops += N * self.dim * self.dim
#         return flops


# class SwinTransformerBlock(nn.Module):
#     r""" Swin Transformer Block.

#     Args:
#         dim (int): Number of input channels.
#         input_resolution (tuple[int]): Input resulotion.
#         num_heads (int): Number of attention heads.
#         window_size (int): Window size.
#         shift_size (int): Shift size for SW-MSA.
#         mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
#         qkv_bias (bool, optional): If True, add a learnable bias to query, key, value. Default: True
#         qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set.
#         drop (float, optional): Dropout rate. Default: 0.0
#         attn_drop (float, optional): Attention dropout rate. Default: 0.0
#         drop_path (float, optional): Stochastic depth rate. Default: 0.0
#         act_layer (nn.Module, optional): Activation layer. Default: nn.GELU
#         norm_layer (nn.Module, optional): Normalization layer.  Default: nn.LayerNorm
#     """

#     def __init__(self, dim, input_resolution, num_heads, window_size=7, shift_size=0,
#                  mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0., drop_path=0.,
#                  act_layer=nn.GELU, norm_layer=nn.LayerNorm):
#         super().__init__()
#         self.dim = dim
#         self.input_resolution = input_resolution
#         self.num_heads = num_heads
#         self.window_size = window_size
#         self.shift_size = shift_size
#         self.mlp_ratio = mlp_ratio
#         if min(self.input_resolution) <= self.window_size:
#             # if window size is larger than input resolution, we don't partition windows
#             self.shift_size = 0
#             self.window_size = min(self.input_resolution)
#         assert 0 <= self.shift_size < self.window_size, "shift_size must in 0-window_size"

#         self.norm1 = norm_layer(dim)
#         self.attn = WindowAttention(
#             dim, window_size=to_2tuple(self.window_size), num_heads=num_heads,
#             qkv_bias=qkv_bias, qk_scale=qk_scale, attn_drop=attn_drop, proj_drop=drop)

#         self.drop_path = DropPath(drop_path) if drop_path > 0. else nn.Identity()
#         self.norm2 = norm_layer(dim)
#         mlp_hidden_dim = int(dim * mlp_ratio)
#         self.mlp = Mlp(in_features=dim, hidden_features=mlp_hidden_dim, act_layer=act_layer, drop=drop)

#         if self.shift_size > 0:
#             attn_mask = self.calculate_mask(self.input_resolution)
#         else:
#             attn_mask = None

#         self.register_buffer("attn_mask", attn_mask)

#     def calculate_mask(self, x_size):
#         # calculate attention mask for SW-MSA
#         H, W = x_size
#         img_mask = torch.zeros((1, H, W, 1))  # 1 H W 1
#         h_slices = (slice(0, -self.window_size),
#                     slice(-self.window_size, -self.shift_size),
#                     slice(-self.shift_size, None))
#         w_slices = (slice(0, -self.window_size),
#                     slice(-self.window_size, -self.shift_size),
#                     slice(-self.shift_size, None))
#         cnt = 0
#         for h in h_slices:
#             for w in w_slices:
#                 img_mask[:, h, w, :] = cnt
#                 cnt += 1

#         mask_windows = window_partition(img_mask, self.window_size)  # nW, window_size, window_size, 1
#         mask_windows = mask_windows.view(-1, self.window_size * self.window_size)
#         attn_mask = mask_windows.unsqueeze(1) - mask_windows.unsqueeze(2)
#         attn_mask = attn_mask.masked_fill(attn_mask != 0, float(-100.0)).masked_fill(attn_mask == 0, float(0.0))

#         return attn_mask

#     def forward(self, x, x_size):
#         H, W = x_size
#         B, L, C = x.shape
#         # assert L == H * W, "input feature has wrong size"

#         shortcut = x
#         x = self.norm1(x)
#         x = x.view(B, H, W, C)

#         # cyclic shift
#         if self.shift_size > 0:
#             shifted_x = torch.roll(x, shifts=(-self.shift_size, -self.shift_size), dims=(1, 2))
#         else:
#             shifted_x = x

#         # partition windows
#         x_windows = window_partition(shifted_x, self.window_size)  # nW*B, window_size, window_size, C
#         x_windows = x_windows.view(-1, self.window_size * self.window_size, C)  # nW*B, window_size*window_size, C

#         # W-MSA/SW-MSA (to be compatible for testing on images whose shapes are the multiple of window size
#         if self.input_resolution == x_size:
#             attn_windows = self.attn(x_windows, mask=self.attn_mask)  # nW*B, window_size*window_size, C
#         else:
#             attn_windows = self.attn(x_windows, mask=self.calculate_mask(x_size).to(x.device))

#         # merge windows
#         attn_windows = attn_windows.view(-1, self.window_size, self.window_size, C)
#         shifted_x = window_reverse(attn_windows, self.window_size, H, W)  # B H' W' C

#         # reverse cyclic shift
#         if self.shift_size > 0:
#             x = torch.roll(shifted_x, shifts=(self.shift_size, self.shift_size), dims=(1, 2))
#         else:
#             x = shifted_x
#         x = x.view(B, H * W, C)

#         # FFN
#         x = shortcut + self.drop_path(x)
#         x = x + self.drop_path(self.mlp(self.norm2(x)))

#         return x

#     def extra_repr(self) -> str:
#         return f"dim={self.dim}, input_resolution={self.input_resolution}, num_heads={self.num_heads}, " \
#                f"window_size={self.window_size}, shift_size={self.shift_size}, mlp_ratio={self.mlp_ratio}"

#     def flops(self):
#         flops = 0
#         H, W = self.input_resolution
#         # norm1
#         flops += self.dim * H * W
#         # W-MSA/SW-MSA
#         nW = H * W / self.window_size / self.window_size
#         flops += nW * self.attn.flops(self.window_size * self.window_size)
#         # mlp
#         flops += 2 * H * W * self.dim * self.dim * self.mlp_ratio
#         # norm2
#         flops += self.dim * H * W
#         return flops


# class PatchMerging(nn.Module):
#     r""" Patch Merging Layer.

#     Args:
#         input_resolution (tuple[int]): Resolution of input feature.
#         dim (int): Number of input channels.
#         norm_layer (nn.Module, optional): Normalization layer.  Default: nn.LayerNorm
#     """

#     def __init__(self, input_resolution, dim, norm_layer=nn.LayerNorm):
#         super().__init__()
#         self.input_resolution = input_resolution
#         self.dim = dim
#         self.reduction = nn.Linear(4 * dim, 2 * dim, bias=False)
#         self.norm = norm_layer(4 * dim)

#     def forward(self, x):
#         """
#         x: B, H*W, C
#         """
#         H, W = self.input_resolution
#         B, L, C = x.shape
#         assert L == H * W, "input feature has wrong size"
#         assert H % 2 == 0 and W % 2 == 0, f"x size ({H}*{W}) are not even."

#         x = x.view(B, H, W, C)

#         x0 = x[:, 0::2, 0::2, :]  # B H/2 W/2 C
#         x1 = x[:, 1::2, 0::2, :]  # B H/2 W/2 C
#         x2 = x[:, 0::2, 1::2, :]  # B H/2 W/2 C
#         x3 = x[:, 1::2, 1::2, :]  # B H/2 W/2 C
#         x = torch.cat([x0, x1, x2, x3], -1)  # B H/2 W/2 4*C
#         x = x.view(B, -1, 4 * C)  # B H/2*W/2 4*C

#         x = self.norm(x)
#         x = self.reduction(x)

#         return x

#     def extra_repr(self) -> str:
#         return f"input_resolution={self.input_resolution}, dim={self.dim}"

#     def flops(self):
#         H, W = self.input_resolution
#         flops = H * W * self.dim
#         flops += (H // 2) * (W // 2) * 4 * self.dim * 2 * self.dim
#         return flops


# class BasicLayer(nn.Module):
#     """ A basic Swin Transformer layer for one stage.

#     Args:
#         dim (int): Number of input channels.
#         input_resolution (tuple[int]): Input resolution.
#         depth (int): Number of blocks.
#         num_heads (int): Number of attention heads.
#         window_size (int): Local window size.
#         mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
#         qkv_bias (bool, optional): If True, add a learnable bias to query, key, value. Default: True
#         qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set.
#         drop (float, optional): Dropout rate. Default: 0.0
#         attn_drop (float, optional): Attention dropout rate. Default: 0.0
#         drop_path (float | tuple[float], optional): Stochastic depth rate. Default: 0.0
#         norm_layer (nn.Module, optional): Normalization layer. Default: nn.LayerNorm
#         downsample (nn.Module | None, optional): Downsample layer at the end of the layer. Default: None
#         use_checkpoint (bool): Whether to use checkpointing to save memory. Default: False.
#     """

#     def __init__(self, dim, input_resolution, depth, num_heads, window_size,
#                  mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0.,
#                  drop_path=0., norm_layer=nn.LayerNorm, downsample=None, use_checkpoint=False):

#         super().__init__()
#         self.dim = dim
#         self.input_resolution = input_resolution
#         self.depth = depth
#         self.use_checkpoint = use_checkpoint

#         # build blocks
#         self.blocks = nn.ModuleList([
#             SwinTransformerBlock(dim=dim, input_resolution=input_resolution,
#                                  num_heads=num_heads, window_size=window_size,
#                                  shift_size=0 if (i % 2 == 0) else window_size // 2,
#                                  mlp_ratio=mlp_ratio,
#                                  qkv_bias=qkv_bias, qk_scale=qk_scale,
#                                  drop=drop, attn_drop=attn_drop,
#                                  drop_path=drop_path[i] if isinstance(drop_path, list) else drop_path,
#                                  norm_layer=norm_layer)
#             for i in range(depth)])

#         # patch merging layer
#         if downsample is not None:
#             self.downsample = downsample(input_resolution, dim=dim, norm_layer=norm_layer)
#         else:
#             self.downsample = None

#     def forward(self, x, x_size):
#         for blk in self.blocks:
#             if self.use_checkpoint:
#                 x = checkpoint.checkpoint(blk, x, x_size)
#             else:
#                 x = blk(x, x_size)
#         if self.downsample is not None:
#             x = self.downsample(x)
#         return x

#     def extra_repr(self) -> str:
#         return f"dim={self.dim}, input_resolution={self.input_resolution}, depth={self.depth}"

#     def flops(self):
#         flops = 0
#         for blk in self.blocks:
#             flops += blk.flops()
#         if self.downsample is not None:
#             flops += self.downsample.flops()
#         return flops


# class RSTB(nn.Module):
#     """Residual Swin Transformer Block (RSTB).

#     Args:
#         dim (int): Number of input channels.
#         input_resolution (tuple[int]): Input resolution.
#         depth (int): Number of blocks.
#         num_heads (int): Number of attention heads.
#         window_size (int): Local window size.
#         mlp_ratio (float): Ratio of mlp hidden dim to embedding dim.
#         qkv_bias (bool, optional): If True, add a learnable bias to query, key, value. Default: True
#         qk_scale (float | None, optional): Override default qk scale of head_dim ** -0.5 if set.
#         drop (float, optional): Dropout rate. Default: 0.0
#         attn_drop (float, optional): Attention dropout rate. Default: 0.0
#         drop_path (float | tuple[float], optional): Stochastic depth rate. Default: 0.0
#         norm_layer (nn.Module, optional): Normalization layer. Default: nn.LayerNorm
#         downsample (nn.Module | None, optional): Downsample layer at the end of the layer. Default: None
#         use_checkpoint (bool): Whether to use checkpointing to save memory. Default: False.
#         img_size: Input image size.
#         patch_size: Patch size.
#         resi_connection: The convolutional block before residual connection.
#     """

#     def __init__(self, dim, input_resolution, depth, num_heads, window_size,
#                  mlp_ratio=4., qkv_bias=True, qk_scale=None, drop=0., attn_drop=0.,
#                  drop_path=0., norm_layer=nn.LayerNorm, downsample=None, use_checkpoint=False,
#                  img_size=224, patch_size=4, resi_connection='1conv'):
#         super(RSTB, self).__init__()

#         self.dim = dim
#         self.input_resolution = input_resolution

#         self.residual_group = BasicLayer(dim=dim,
#                                          input_resolution=input_resolution,
#                                          depth=depth,
#                                          num_heads=num_heads,
#                                          window_size=window_size,
#                                          mlp_ratio=mlp_ratio,
#                                          qkv_bias=qkv_bias, qk_scale=qk_scale,
#                                          drop=drop, attn_drop=attn_drop,
#                                          drop_path=drop_path,
#                                          norm_layer=norm_layer,
#                                          downsample=downsample,
#                                          use_checkpoint=use_checkpoint)

#         if resi_connection == '1conv':
#             self.conv = nn.Conv2d(dim, dim, 3, 1, 1)
#         elif resi_connection == '3conv':
#             # to save parameters and memory
#             self.conv = nn.Sequential(nn.Conv2d(dim, dim // 4, 3, 1, 1), nn.LeakyReLU(negative_slope=0.2, inplace=True),
#                                       nn.Conv2d(dim // 4, dim // 4, 1, 1, 0),
#                                       nn.LeakyReLU(negative_slope=0.2, inplace=True),
#                                       nn.Conv2d(dim // 4, dim, 3, 1, 1))

#         self.patch_embed = PatchEmbed(
#             img_size=img_size, patch_size=patch_size, in_chans=0, embed_dim=dim,
#             norm_layer=None)

#         self.patch_unembed = PatchUnEmbed(
#             img_size=img_size, patch_size=patch_size, in_chans=0, embed_dim=dim,
#             norm_layer=None)

#     def forward(self, x, x_size):
#         return self.patch_embed(self.conv(self.patch_unembed(self.residual_group(x, x_size), x_size))) + x

#     def flops(self):
#         flops = 0
#         flops += self.residual_group.flops()
#         H, W = self.input_resolution
#         flops += H * W * self.dim * self.dim * 9
#         flops += self.patch_embed.flops()
#         flops += self.patch_unembed.flops()

#         return flops


# class PatchEmbed(nn.Module):
#     r""" Image to Patch Embedding

#     Args:
#         img_size (int): Image size.  Default: 224.
#         patch_size (int): Patch token size. Default: 4.
#         in_chans (int): Number of input image channels. Default: 3.
#         embed_dim (int): Number of linear projection output channels. Default: 96.
#         norm_layer (nn.Module, optional): Normalization layer. Default: None
#     """

#     def __init__(self, img_size=224, patch_size=4, in_chans=3, embed_dim=96, norm_layer=None):
#         super().__init__()
#         img_size = to_2tuple(img_size)
#         patch_size = to_2tuple(patch_size)
#         patches_resolution = [img_size[0] // patch_size[0], img_size[1] // patch_size[1]]
#         self.img_size = img_size
#         self.patch_size = patch_size
#         self.patches_resolution = patches_resolution
#         self.num_patches = patches_resolution[0] * patches_resolution[1]

#         self.in_chans = in_chans
#         self.embed_dim = embed_dim

#         if norm_layer is not None:
#             self.norm = norm_layer(embed_dim)
#         else:
#             self.norm = None

#     def forward(self, x):
#         x = x.flatten(2).transpose(1, 2)  # B Ph*Pw C
#         if self.norm is not None:
#             x = self.norm(x)
#         return x

#     def flops(self):
#         flops = 0
#         H, W = self.img_size
#         if self.norm is not None:
#             flops += H * W * self.embed_dim
#         return flops


# class PatchUnEmbed(nn.Module):
#     r""" Image to Patch Unembedding

#     Args:
#         img_size (int): Image size.  Default: 224.
#         patch_size (int): Patch token size. Default: 4.
#         in_chans (int): Number of input image channels. Default: 3.
#         embed_dim (int): Number of linear projection output channels. Default: 96.
#         norm_layer (nn.Module, optional): Normalization layer. Default: None
#     """

#     def __init__(self, img_size=224, patch_size=4, in_chans=3, embed_dim=96, norm_layer=None):
#         super().__init__()
#         img_size = to_2tuple(img_size)
#         patch_size = to_2tuple(patch_size)
#         patches_resolution = [img_size[0] // patch_size[0], img_size[1] // patch_size[1]]
#         self.img_size = img_size
#         self.patch_size = patch_size
#         self.patches_resolution = patches_resolution
#         self.num_patches = patches_resolution[0] * patches_resolution[1]

#         self.in_chans = in_chans
#         self.embed_dim = embed_dim

#     def forward(self, x, x_size):
#         B, HW, C = x.shape
#         x = x.transpose(1, 2).view(B, self.embed_dim, x_size[0], x_size[1])  # B Ph*Pw C
#         return x

#     def flops(self):
#         flops = 0
#         return flops


# class Upsample(nn.Sequential):
#     """Upsample module.

#     Args:
#         scale (int): Scale factor. Supported scales: 2^n and 3.
#         num_feat (int): Channel number of intermediate features.
#     """

#     def __init__(self, scale, num_feat):
#         m = []
#         if (scale & (scale - 1)) == 0:  # scale = 2^n
#             for _ in range(int(math.log(scale, 2))):
#                 m.append(nn.Conv2d(num_feat, 4 * num_feat, 3, 1, 1))
#                 m.append(nn.PixelShuffle(2))
#         elif scale == 3:
#             m.append(nn.Conv2d(num_feat, 9 * num_feat, 3, 1, 1))
#             m.append(nn.PixelShuffle(3))
#         else:
#             raise ValueError(f'scale {scale} is not supported. ' 'Supported scales: 2^n and 3.')
#         super(Upsample, self).__init__(*m)


# class UpsampleOneStep(nn.Sequential):
#     """UpsampleOneStep module (the difference with Upsample is that it always only has 1conv + 1pixelshuffle)
#        Used in lightweight SR to save parameters.

#     Args:
#         scale (int): Scale factor. Supported scales: 2^n and 3.
#         num_feat (int): Channel number of intermediate features.

#     """

#     def __init__(self, scale, num_feat, num_out_ch, input_resolution=None):
#         self.num_feat = num_feat
#         self.input_resolution = input_resolution
#         m = []
#         m.append(nn.Conv2d(num_feat, (scale ** 2) * num_out_ch, 3, 1, 1))
#         m.append(nn.PixelShuffle(scale))
#         super(UpsampleOneStep, self).__init__(*m)

#     def flops(self):
#         H, W = self.input_resolution
#         flops = H * W * self.num_feat * 3 * 9
#         return flops


# class SwinIR(nn.Module):
#     r""" SwinIR
#         A PyTorch impl of : `SwinIR: Image Restoration Using Swin Transformer`, based on Swin Transformer.

#     Args:
#         img_size (int | tuple(int)): Input image size. Default 64
#         patch_size (int | tuple(int)): Patch size. Default: 1
#         in_chans (int): Number of input image channels. Default: 3
#         embed_dim (int): Patch embedding dimension. Default: 96
#         depths (tuple(int)): Depth of each Swin Transformer layer.
#         num_heads (tuple(int)): Number of attention heads in different layers.
#         window_size (int): Window size. Default: 7
#         mlp_ratio (float): Ratio of mlp hidden dim to embedding dim. Default: 4
#         qkv_bias (bool): If True, add a learnable bias to query, key, value. Default: True
#         qk_scale (float): Override default qk scale of head_dim ** -0.5 if set. Default: None
#         drop_rate (float): Dropout rate. Default: 0
#         attn_drop_rate (float): Attention dropout rate. Default: 0
#         drop_path_rate (float): Stochastic depth rate. Default: 0.1
#         norm_layer (nn.Module): Normalization layer. Default: nn.LayerNorm.
#         ape (bool): If True, add absolute position embedding to the patch embedding. Default: False
#         patch_norm (bool): If True, add normalization after patch embedding. Default: True
#         use_checkpoint (bool): Whether to use checkpointing to save memory. Default: False
#         upscale: Upscale factor. 2/3/4/8 for image SR, 1 for denoising and compress artifact reduction
#         img_range: Image range. 1. or 255.
#         upsampler: The reconstruction reconstruction module. 'pixelshuffle'/'pixelshuffledirect'/'nearest+conv'/None
#         resi_connection: The convolutional block before residual connection. '1conv'/'3conv'
#     """

#     def __init__(self, img_size=128, patch_size=1, in_chans=1,
#                  embed_dim=180, depths=[6, 6, 6, 6, 6, 6], num_heads=[6, 6, 6, 6, 6, 6],
#                  window_size=8, mlp_ratio=2., qkv_bias=True, qk_scale=None,
#                  drop_rate=0., attn_drop_rate=0., drop_path_rate=0.1,
#                  norm_layer=nn.LayerNorm, ape=False, patch_norm=True,
#                  use_checkpoint=False, upscale=2, img_range=1., upsampler='none', resi_connection='1conv',
#                  **kwargs):
#         super(SwinIR, self).__init__()
#         num_in_ch = in_chans
#         num_out_ch = in_chans
#         num_feat = 64
#         self.img_range = img_range
#         if in_chans == 3:
#             rgb_mean = (0.4488, 0.4371, 0.4040)
#             self.mean = torch.Tensor(rgb_mean).view(1, 3, 1, 1)
#         else:
#             self.mean = torch.zeros(1, 1, 1, 1)
#         self.upscale = upscale
#         self.upsampler = upsampler
#         self.window_size = window_size
#         self.out_dim = num_feat
#         #####################################################################################################
#         ################################### 1, shallow feature extraction ###################################
#         self.conv_first = nn.Conv2d(num_in_ch, embed_dim, 3, 1, 1)

#         #####################################################################################################
#         ################################### 2, deep feature extraction ######################################
#         self.num_layers = len(depths)
#         self.embed_dim = embed_dim
#         self.ape = ape
#         self.patch_norm = patch_norm
#         self.num_features = embed_dim
#         self.mlp_ratio = mlp_ratio

#         # split image into non-overlapping patches
#         self.patch_embed = PatchEmbed(
#             img_size=img_size, patch_size=patch_size, in_chans=embed_dim, embed_dim=embed_dim,
#             norm_layer=norm_layer if self.patch_norm else None)
#         num_patches = self.patch_embed.num_patches
#         patches_resolution = self.patch_embed.patches_resolution
#         self.patches_resolution = patches_resolution

#         # merge non-overlapping patches into image
#         self.patch_unembed = PatchUnEmbed(
#             img_size=img_size, patch_size=patch_size, in_chans=embed_dim, embed_dim=embed_dim,
#             norm_layer=norm_layer if self.patch_norm else None)

#         # absolute position embedding
#         if self.ape:
#             self.absolute_pos_embed = nn.Parameter(torch.zeros(1, num_patches, embed_dim))
#             trunc_normal_(self.absolute_pos_embed, std=.02)

#         self.pos_drop = nn.Dropout(p=drop_rate)

#         # stochastic depth
#         dpr = [x.item() for x in torch.linspace(0, drop_path_rate, sum(depths))]  # stochastic depth decay rule

#         # build Residual Swin Transformer blocks (RSTB)
#         self.layers = nn.ModuleList()
#         for i_layer in range(self.num_layers):
#             layer = RSTB(dim=embed_dim,
#                          input_resolution=(patches_resolution[0],
#                                            patches_resolution[1]),
#                          depth=depths[i_layer],
#                          num_heads=num_heads[i_layer],
#                          window_size=window_size,
#                          mlp_ratio=self.mlp_ratio,
#                          qkv_bias=qkv_bias, qk_scale=qk_scale,
#                          drop=drop_rate, attn_drop=attn_drop_rate,
#                          drop_path=dpr[sum(depths[:i_layer]):sum(depths[:i_layer + 1])],  # no impact on SR results
#                          norm_layer=norm_layer,
#                          downsample=None,
#                          use_checkpoint=use_checkpoint,
#                          img_size=img_size,
#                          patch_size=patch_size,
#                          resi_connection=resi_connection

#                          )
#             self.layers.append(layer)
#         self.norm = norm_layer(self.num_features)

#         # build the last conv layer in deep feature extraction
#         if resi_connection == '1conv':
#             self.conv_after_body = nn.Conv2d(embed_dim, embed_dim, 3, 1, 1)
#         elif resi_connection == '3conv':
#             # to save parameters and memory
#             self.conv_after_body = nn.Sequential(nn.Conv2d(embed_dim, embed_dim // 4, 3, 1, 1),
#                                                  nn.LeakyReLU(negative_slope=0.2, inplace=True),
#                                                  nn.Conv2d(embed_dim // 4, embed_dim // 4, 1, 1, 0),
#                                                  nn.LeakyReLU(negative_slope=0.2, inplace=True),
#                                                  nn.Conv2d(embed_dim // 4, embed_dim, 3, 1, 1))

#         #####################################################################################################
#         ################################ 3, high quality image reconstruction ################################
#         if self.upsampler == 'none':
#             self.conv_before_upsample = nn.Sequential(nn.Conv2d(embed_dim, num_feat, 3, 1, 1),
#                                                       nn.LeakyReLU(inplace=True))        
#         elif self.upsampler == 'pixelshuffle':
#             # for classical SR
#             self.conv_before_upsample = nn.Sequential(nn.Conv2d(embed_dim, num_feat, 3, 1, 1),
#                                                       nn.LeakyReLU(inplace=True))
#             self.upsample = Upsample(upscale, num_feat)
#             self.conv_last = nn.Conv2d(num_feat, num_out_ch, 3, 1, 1)
#         elif self.upsampler == 'pixelshuffledirect':
#             # for lightweight SR (to save parameters)
#             self.upsample = UpsampleOneStep(upscale, embed_dim, num_out_ch,
#                                             (patches_resolution[0], patches_resolution[1]))
#         elif self.upsampler == 'nearest+conv':
#             # for real-world SR (less artifacts)
#             assert self.upscale == 4, 'only support x4 now.'
#             self.conv_before_upsample = nn.Sequential(nn.Conv2d(embed_dim, num_feat, 3, 1, 1),
#                                                       nn.LeakyReLU(inplace=True))
#             self.conv_up1 = nn.Conv2d(num_feat, num_feat, 3, 1, 1)
#             self.conv_up2 = nn.Conv2d(num_feat, num_feat, 3, 1, 1)
#             self.conv_hr = nn.Conv2d(num_feat, num_feat, 3, 1, 1)
#             self.conv_last = nn.Conv2d(num_feat, num_out_ch, 3, 1, 1)
#             self.lrelu = nn.LeakyReLU(negative_slope=0.2, inplace=True)
#         else:
#             # for image denoising and JPEG compression artifact reduction
#             self.conv_last = nn.Conv2d(embed_dim, num_out_ch, 3, 1, 1)

#         self.apply(self._init_weights)

#     def _init_weights(self, m):
#         if isinstance(m, nn.Linear):
#             trunc_normal_(m.weight, std=.02)
#             if isinstance(m, nn.Linear) and m.bias is not None:
#                 nn.init.constant_(m.bias, 0)
#         elif isinstance(m, nn.LayerNorm):
#             nn.init.constant_(m.bias, 0)
#             nn.init.constant_(m.weight, 1.0)

#     @torch.jit.ignore
#     def no_weight_decay(self):
#         return {'absolute_pos_embed'}

#     @torch.jit.ignore
#     def no_weight_decay_keywords(self):
#         return {'relative_position_bias_table'}

#     def check_image_size(self, x):
#         _, _, h, w = x.size()
#         mod_pad_h = (self.window_size - h % self.window_size) % self.window_size
#         mod_pad_w = (self.window_size - w % self.window_size) % self.window_size
#         x = F.pad(x, (0, mod_pad_w, 0, mod_pad_h), 'reflect')
#         return x

#     def forward_features(self, x):
#         x_size = (x.shape[2], x.shape[3])
#         x = self.patch_embed(x)
#         if self.ape:
#             x = x + self.absolute_pos_embed
#         x = self.pos_drop(x)

#         for layer in self.layers:
#             x = layer(x, x_size)

#         x = self.norm(x)  # B L C
#         x = self.patch_unembed(x, x_size)

#         return x

#     def forward(self, x):
#         x = self.check_image_size(x)
        
# #         self.mean = self.mean.type_as(x)
# #         x = (x - self.mean) * self.img_range
        
#         if self.upsampler == 'none':
#             x = self.conv_first(x)
#             x = self.conv_after_body(self.forward_features(x)) + x
#             x = self.conv_before_upsample(x)
#         elif self.upsampler == 'pixelshuffle':
#             # for classical SR
#             x = self.conv_first(x)
#             x = self.conv_after_body(self.forward_features(x)) + x
#             x = self.conv_before_upsample(x)
#             x = self.conv_last(self.upsample(x))
#         elif self.upsampler == 'pixelshuffledirect':
#             # for lightweight SR
#             x = self.conv_first(x)
#             x = self.conv_after_body(self.forward_features(x)) + x
#             x = self.upsample(x)
#         elif self.upsampler == 'nearest+conv':
#             # for real-world SR
#             x = self.conv_first(x)
#             x = self.conv_after_body(self.forward_features(x)) + x
#             x = self.conv_before_upsample(x)
#             x = self.lrelu(self.conv_up1(torch.nn.functional.interpolate(x, scale_factor=2, mode='nearest')))
#             x = self.lrelu(self.conv_up2(torch.nn.functional.interpolate(x, scale_factor=2, mode='nearest')))
#             x = self.conv_last(self.lrelu(self.conv_hr(x)))
#         else:
#             # for image denoising and JPEG compression artifact reduction
#             x_first = self.conv_first(x)
#             res = self.conv_after_body(self.forward_features(x_first)) + x_first
#             x = x + self.conv_last(res)

# #         x = x / self.img_range + self.mean

#         return x

#     def flops(self):
#         flops = 0
#         H, W = self.patches_resolution
#         flops += H * W * 3 * self.embed_dim * 9
#         flops += self.patch_embed.flops()
#         for i, layer in enumerate(self.layers):
#             flops += layer.flops()
#         flops += H * W * 3 * self.embed_dim * self.embed_dim
#         flops += self.upsample.flops()
#         return flops

# def make_swinir(no_upsampling=True):
#     return SwinIR()


# based on WIRE: Wavelet Implicit Neural Representations
# Saragadam et al. (CVPR 2023)
# Code adapted from: https://github.com/vishwa91/wire

import torch
import torch.nn as nn

class ComplexGaborLayer(nn.Module):
    '''
        Implicit representation with complex Gabor nonlinearity
        
        Inputs;
            in_features: Input features
            out_features; Output features
            bias: if True, enable bias for the linear operation
            is_first: Legacy SIREN parameter
            omega_0: Legacy SIREN parameter
            omega0: Frequency of Gabor sinusoid term
            sigma0: Scaling of Gabor Gaussian term
            trainable: If True, omega and sigma are trainable parameters
    '''
    
    def __init__(self, in_features, out_features, bias=True,
                 is_first=False, omega0=10.0, sigma0=40.0,
                 trainable=True):
        super().__init__()
        self.omega_0 = omega0
        self.scale_0 = sigma0
        self.is_first = is_first
        
        self.in_features = in_features
        
        if self.is_first:
            dtype = torch.float
        else:
            dtype = torch.cfloat
            
        # Set trainable parameters if they are to be simultaneously optimized
        self.omega_0 = nn.Parameter(self.omega_0*torch.ones(1), trainable)
        self.scale_0 = nn.Parameter(self.scale_0*torch.ones(1), trainable)
        
        self.linear = nn.Linear(in_features,
                                out_features,
                                bias=bias,
                                dtype=dtype)
    
    def forward(self, input):
        lin = self.linear(input)
        omega = self.omega_0 * lin
        scale = self.scale_0 * lin
        
        return torch.exp(1j*omega - scale.abs().square())


import torch
from torch import nn
from niiv.feature_grid import FeatureGrid
from niiv.encoders.edsr_2d import EDSR2D
from niiv.decoder.mlp import MLP
from niiv.encoders import rdn
from niiv.encoders import swinir
from niiv.decoder import inr
from niiv.decoder.field_siren import FieldSiren

class NIIV(nn.Module):
    def __init__(self, out_features=1, encoding_config=None, n_pos_enc_octaves=2, **kwargs):
        super().__init__()

        self.feat_unfold = False

        if encoding_config is None:
            n_features = 64
            n_layers = 5
            n_neurons = 256
            self.encoder = EDSR2D()
        else:
            # hyper parameters
            n_features = encoding_config["encoder"]["n_features"]
            n_layers = encoding_config["network"]["n_hidden_layers"]
            n_neurons = encoding_config["network"]["n_neurons"]
            self.encoder = EDSR2D(args = encoding_config["encoder"])   

        # module for latent grid processing
        self.grid = FeatureGrid(feat_unfold=self.feat_unfold, n_pos_encoding=n_pos_enc_octaves)
        model_in = self.grid.n_out(n_features) 

        # trainable parameters
        # self.encoder = rdn.make_rdn()
        # self.encoder = swinir.make_swinir()
        self.decoder = MLP(in_dim=model_in, out_dim=out_features, n_neurons=n_neurons, n_hidden=n_layers)

    def forward(self, image, coords):
        latent_grid = self.encoder(image)
        features = self.grid.compute_features(image, latent_grid, coords)
        bs, q = coords.squeeze(1).squeeze(1).shape[:2]
        prediction = self.decoder(features.view(bs * q, -1)).view(bs, q, -1)
        return torch.clamp(prediction, 0, 1)



from setuptools import setup, find_packages

setup(
    name='niiv',
    version='0.0.5',
    packages=find_packages(),
    install_requires=[
        'torch',
        'DISTS_pytorch',
        'lpips',
        'numpy',
        'pytorch-ignite',
        'tqdm',
        'wandb',
        'pytorch_msssim'
    ],
    # Other metadata
    author='Anonymous Anonymous',
    author_email='anonymous@anonymous.com',
    description='Self-Supervised Neural Implicit Isotropic Volume Reconstruction',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',  # This is important for a proper display on PyPI
    url='https://niiv.net',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torchmetrics.functional.image import image_gradients
from util.transforms import crop_image_border
import matplotlib.pyplot as plt
from util.eval_metrics import FourierDenoiser


class GradientRegularizer(nn.Module):
    def __init__(self):
        super(GradientRegularizer, self).__init__()
        self.sobel_x = torch.tensor([[-0.25, 0, 0.25], [-0.5, 0, 0.5], [-0.25, 0, 0.25]], dtype=torch.float32).view(1, 1, 3, 3).cuda()
        self.sobel_y = torch.tensor([[-0.25, -0.5, -0.25], [0, 0, 0], [0.25, 0.5, 0.25]], dtype=torch.float32).view(1, 1, 3, 3).cuda()
        self.denoiser = FourierDenoiser(threshold=25)


    def forward(self, x, epoch, step, weight=0.01):
        x = self.denoiser(x)
        x = x.unsqueeze(1)
        # dy, dx = image_gradients(x)

        x_padded = F.pad(x, (1, 1, 1, 1), mode='replicate')

        G_x = F.conv2d(x_padded, self.sobel_x)
        G_y = F.conv2d(x_padded, self.sobel_y)

        mag = torch.sqrt(G_x**2 + G_y**2 + 1e-9)
        # mag = mag - 0.05
        # mag = torch.clamp(mag, 0.0, 0.35) # disallow noise like and large gradients to dominate the loss
        # mag = crop_image_border(mag, 5)


        # if epoch % 5 == 0 and step == 0 :
        #     ## plot histogram of dy, dx and mag
        #     plt.hist(dy.flatten().detach().cpu().numpy(), bins=100)
        #     plt.savefig('histograms/dy-histogram-{}.png'.format(epoch)) 
        #     plt.clf()   
        #     plt.hist(dx.flatten().detach().cpu().numpy(), bins=100)
        #     plt.savefig('histograms/dx-histogram-{}.png'.format(epoch))
        #     plt.clf()
        #     plt.hist(mag.flatten().detach().cpu().numpy(), bins=100)
        #     plt.savefig('histograms/mag-histogram-{}.png'.format(epoch))
        #     plt.clf()

        x = torch.pow(1.0 - torch.mean(torch.abs(mag)), 1)
        x = torch.clamp(x, 0.0, 1.0)
        return x * weight


class FourierRegularizer(nn.Module):
    def __init__(self, filter_radius=25):
        super(FourierRegularizer, self).__init__()
        self.filter_radius = filter_radius

    def forward(self, x):
        # Compute the 2D Fourier Transform of the image
        f_transform = torch.fft.fft2(x)
        f_shifted = torch.fft.fftshift(f_transform)

        # Create a low-pass filter mask (circular mask)
        rows, cols = x.shape[-2:]
        crow, ccol = rows // 2 , cols // 2
        low_pass = torch.zeros((rows, cols), dtype=torch.uint8)
        y, x = torch.meshgrid[:rows, :cols]
        mask_area = (x - ccol) ** 2 + (y - crow) ** 2 <= self.filter_radius**2
        low_pass[mask_area] = 1

        # Apply the mask/filter
        f_shifted_filtered = f_shifted * low_pass

        # Inverse Fourier Transform to get the denoised image back
        f_ishifted = torch.fft.ifftshift(f_shifted_filtered)
        img_back = torch.fft.ifft2(f_ishifted)
        img_back = torch.abs(img_back)
        return img_back



import torch
from niiv.util.utils import make_coord
from niiv.decoder.pos_enc import PositionalEncoding

class FeatureGrid():
    def __init__(self, feat_unfold, n_pos_encoding):
        self.feature_unfold = feat_unfold
        self.pos_enc = PositionalEncoding(num_octaves=n_pos_encoding)
    
    def n_out(self, n_in):
        if self.feature_unfold:
            return n_in * 9
        return n_in + self.pos_enc.d_out(2) + 1
        # return n_in + 2 + 1
    
    def compute_features(self, image, latents, coords):

        if self.feature_unfold:
            # concat each latent by it's local neighborhood
            latents = self.unfold_features(latents)

        # interpolate feature coordinates
        feature_coords = make_coord(latents.shape[-2:], flatten=False).cuda()
        feature_coords = feature_coords.permute(2, 0, 1).unsqueeze(0)
        feature_coords = feature_coords.repeat(coords.shape[0], 1, 1, 1)

        coords_ = coords.unsqueeze(1)

        # interpolate features
        q_features = torch.nn.functional.grid_sample(latents, coords_.flip(-1), mode='bilinear', align_corners=None)
        q_features = q_features.squeeze(2).squeeze(2)
        q_features = q_features.permute(0, 2, 1)

        q_coords = torch.nn.functional.grid_sample(feature_coords, coords_.flip(-1), mode='bilinear', align_corners=None)
        q_coords = q_coords.squeeze(2).squeeze(2)
        q_coords = q_coords.permute(0, 2, 1)

        q_input = torch.nn.functional.grid_sample(image, coords_.flip(-1), mode='bilinear', align_corners=None)
        q_input = q_input.squeeze(2).squeeze(2)
        q_input = q_input.permute(0, 2, 1)

        pe_coords = self.pos_enc((q_coords + 1.0) / 2.0)

        # pe_coords = q_coords

        decoder_input = torch.cat((q_features, q_input, pe_coords.squeeze()), dim=-1)
        return decoder_input
    
    def unfold_features(self, latents):
        unfold_list = [-1, 0, 1]
        bs, n_feat, x_dim, y_dim = latents.shape

        x_idx = torch.arange(0, latents.shape[-2], dtype=torch.int64).cuda()
        y_idx = torch.arange(0, latents.shape[-1], dtype=torch.int64).cuda()

        tmp_x, tmp_y = torch.meshgrid(x_idx, y_idx)

        idx = torch.stack((tmp_x, tmp_y), dim=-1)

        x = idx[..., 0].flatten()
        y = idx[..., 1].flatten()

        neighbors = []

        for i in unfold_list:
            for j in unfold_list:
                vx = torch.clamp(x+i, 0, latents.shape[-2]-1)
                vy = torch.clamp(y+j, 0, latents.shape[-1]-1)

                nbh = latents[:, :, vx, vy]
                neighbors.append(nbh)

        latents = torch.stack(neighbors, dim=1).view(bs, n_feat * len(unfold_list)**2, x_dim, y_dim)
        return latents